@@ -10,6 +10,7 @@ from typing import Any, Dict, List, Union
 import numpy as np
 import pandas as pd
 from pandas import DataFrame, Series
+from pandas.errors import ParserError
 from requests import Response
 
 from seclea_ai.authentication import AuthenticationService
@@ -21,9 +22,10 @@ from seclea_ai.lib.seclea_utils.core import (
     encode_func,
     save_object,
 )
+import uuid
+from .lib.seclea_utils.dataset_management.dataset_utils import save_dataset, dataset_hash
 from seclea_ai.lib.seclea_utils.model_management.get_model_manager import ModelManagers, serialize
 from seclea_ai.transformations import DatasetTransformation
-
 from .svc.api.collection.dataset import post_dataset
 from .svc.api.collection.model_state import post_model_state
 
@@ -45,9 +47,7 @@ def handle_response(res: Response, expected: int, msg: str) -> Response:
 
     """
     if not res.status_code == expected:
-        raise ValueError(
-            f"Response Status code {res.status_code}, expected:{expected}. \n{msg} - {res.reason} - {res.text}"
-        )
+        raise ValueError(f"Response Status code {res.status_code}, expected:{expected}. \n{msg}")
     return res
 
 
@@ -206,6 +206,7 @@ class SecleaAI:
             >>> dataset_metadata = {"index": "TransactionID", "outcome_name": "isFraud", "continuous_features": ["TransactionDT", "TransactionAmt"]}
             >>> seclea.upload_dataset(dataset=files, dataset_name="Multifile Dataset", metadata=dataset_metadata)
 
+
         """
         # processing the final dataset - make sure it's a DataFrame
         if self._project is None:
@@ -216,61 +217,41 @@ class SecleaAI:
         elif isinstance(dataset, str):
             dataset = pd.read_csv(dataset, index_col=metadata["index"])
 
-        dataset_hash = pd.util.hash_pandas_object(dataset).sum()
+        dset_pk = dataset_hash(dataset, self._project)
 
         if transformations is not None:
 
-            # check that the parent exists on platform
             parent = self._assemble_dataset(transformations[0].raw_data_kwargs)
 
-            # setup for generating datasets.
-            last = len(transformations) - 1
-            upload_queue = list()
+            #####
+            # Validate parent exists and get metadata - can factor out
+            #####
+            parent_dset_pk = dataset_hash(parent, self._project)
+            # check parent exists - throw an error if not.
+            res = self._transmission.get(
+                url_path=f"/collection/datasets/{parent_dset_pk}",
+                query_params={"project": self._project, "organization": self._organization},
+            )
+            if not res.status_code == 200:
+                raise AssertionError(
+                    "Parent Dataset does not exist on the Platform. Please check your arguments and "
+                    "that you have uploaded the parent dataset already"
+                )
+            parent_metadata = res.json()["metadata"]
+            #####
 
-            output = dict()
-            # iterate over transformations, assembling intermediate datasets
-            # TODO address memory issue of keeping all datasets
-            for idx, trans in enumerate(transformations):
-                output = trans(output)
-
-                # construct the generated dataset from outputs
-                dset = self._assemble_dataset(output)
-                dset_metadata = {}
-                dset_name = f"{dataset_name}-{trans.func.__name__}"
-
-                # handle the final dataset - check generated = passed in.
-                if idx == last:
-                    if (
-                        pd.util.hash_pandas_object(dset).sum() != dataset_hash
-                    ):  # TODO create or find better exception
-                        raise AssertionError(
-                            """Generated Dataset does not match the Dataset passed in.
-                                         Please check your DatasetTransformation definitions and try again.
-                                         Try using less DatasetTransformations if you are having persistent problems"""
-                        )
-                    else:
-                        dset_metadata = metadata
-                        dset_name = dataset_name
-
-                # add data to queue to upload later after final dataset checked
-                upload_kwargs = {
-                    "dataset": copy.deepcopy(dset),  # TODO change keys
-                    "dataset_name": copy.deepcopy(dset_name),
-                    "metadata": copy.deepcopy(dset_metadata),
-                    "parent_hash": pd.util.hash_pandas_object(parent).sum(),
-                    "transformation": copy.deepcopy(trans),
-                }
-                # update the parent dataset - these chained transformations only make sense if they are pushing the
-                # same dataset through multiple transformations.
-                parent = copy.deepcopy(dset)
-                upload_queue.append(upload_kwargs)
+            upload_queue = self._generate_intermediate_datasets(
+                transformations=transformations,
+                dataset_name=dataset_name,
+                dset_pk=dset_pk,
+                user_metadata=metadata,
+                parent=parent,
+                parent_metadata=parent_metadata,
+            )
 
             # check for duplicates
             for up_kwargs in upload_queue:
-                if (
-                    pd.util.hash_pandas_object(up_kwargs["dataset"]).sum()
-                    == up_kwargs["parent_hash"]
-                ):
+                if dataset_hash(up_kwargs["dataset"], self._project) == up_kwargs["parent_hash"]:
                     raise AssertionError(
                         f"""The transformation {up_kwargs['transformation'].func.__name__} does not change the dataset.
                         Please remove it and try again."""
@@ -281,6 +262,40 @@ class SecleaAI:
             return
 
         # this only happens if this has no transformations ie. it is a Raw Dataset.
+        metadata_defaults_spec = dict(
+            continuous_features=[],
+            outcome_name=None,
+            num_samples=len(dataset),
+            favourable_outcome=None,
+            unfavourable_outcome=None,
+            dataset_type=self._get_dataset_type(dataset),
+        )
+        try:
+            features = (
+                dataset.columns
+            )  # TODO - drop the outcome name but requires changes on frontend.
+        except KeyError:
+            # this means outcome was set to None
+            features = dataset.columns
+
+        metadata = self._ensure_required_metadata(
+            metadata=metadata, defaults_spec=metadata_defaults_spec
+        )
+        automatic_metadata = dict(
+            index=0 if dataset.index.name is None else dataset.index.name,
+            split=None,
+            features=list(features),
+            categorical_features=list(
+                set(list(features))
+                - set(metadata["continuous_features"]).intersection(set(list(features)))
+            ),
+        )
+        metadata = self._add_required_metadata(metadata=metadata, required_spec=automatic_metadata)
+
+        metadata["categorical_values"] = [
+            {col: dataset[col].unique().tolist()} for col in metadata["categorical_features"]
+        ]
+
         self._upload_dataset(
             dataset=dataset,
             dataset_name=dataset_name,
@@ -289,6 +304,101 @@ class SecleaAI:
             transformation=None,
         )
 
+    def _generate_intermediate_datasets(
+        self, transformations, dataset_name, dset_pk, user_metadata, parent, parent_metadata
+    ):
+        # setup for generating datasets.
+        last = len(transformations) - 1
+        upload_queue = list()
+        parent_mdata = parent_metadata
+        parent_dset = parent
+
+        output = dict()
+        # iterate over transformations, assembling intermediate datasets
+        # TODO address memory issue of keeping all datasets
+        for idx, trans in enumerate(transformations):
+            output = trans(output)
+
+            # construct the generated dataset from outputs
+            dset = self._assemble_dataset(output)
+
+            dset_metadata = copy.deepcopy(user_metadata)
+            # validate and ensure required metadata
+            metadata_defaults_spec = dict(
+                continuous_features=parent_mdata["continuous_features"],
+                outcome_name=parent_mdata["outcome_name"],
+                num_samples=len(dset),
+                favourable_outcome=parent_mdata["favourable_outcome"],
+                unfavourable_outcome=parent_mdata["unfavourable_outcome"],
+            )
+            dset_metadata = self._ensure_required_metadata(
+                metadata=dset_metadata, defaults_spec=metadata_defaults_spec
+            )
+            try:
+                features = (
+                    dset.columns
+                )  # TODO - drop the outcome name but requires changes on frontend.
+            except KeyError:
+                # this means outcome was set to None
+                features = dset.columns
+
+            automatic_metadata = dict(
+                index=0 if dset.index.name is None else dset.index.name,
+                split=trans.split if trans is not None else parent_mdata["split"],
+                features=list(features),
+                categorical_features=list(
+                    set(list(features))
+                    - set(dset_metadata["continuous_features"]).intersection(set(list(features)))
+                ),
+            )
+
+            dset_metadata = self._add_required_metadata(
+                metadata=dset_metadata, required_spec=automatic_metadata
+            )
+
+            dset_metadata["categorical_values"] = [
+                {col: dset[col].unique().tolist()} for col in dset_metadata["categorical_features"]
+            ]
+
+            # constraints
+            if not set(dset_metadata["continuous_features"]).issubset(
+                set(dset_metadata["features"])
+            ):
+                raise ValueError(
+                    "Continuous features must be a subset of features. Please check and try again."
+                )
+
+            dset_name = f"{dataset_name}-{trans.func.__name__}"  # TODO improve this.
+
+            # handle the final dataset - check generated = passed in.
+            if idx == last:
+                if (
+                    dataset_hash(dset, self._project) != dset_pk
+                ):  # TODO create or find better exception
+                    raise AssertionError(
+                        """Generated Dataset does not match the Dataset passed in.
+                                     Please check your DatasetTransformation definitions and try again.
+                                     Try using less DatasetTransformations if you are having persistent problems"""
+                    )
+                else:
+                    dset_name = dataset_name
+
+            # add data to queue to upload later after final dataset checked
+            upload_kwargs = {
+                "dataset": copy.deepcopy(dset),  # TODO change keys
+                "dataset_name": copy.deepcopy(dset_name),
+                "metadata": dset_metadata,
+                "parent_hash": dataset_hash(parent_dset, self._project),
+                "transformation": copy.deepcopy(trans),
+            }
+            # update the parent dataset - these chained transformations only make sense if they are pushing the
+            # same dataset through multiple transformations.
+            parent_dset = copy.deepcopy(dset)
+            parent_mdata = copy.deepcopy(dset_metadata)
+            upload_queue.append(upload_kwargs)
+
+        return upload_queue
+
     def upload_training_run_split(
         self,
         model,
@@ -368,7 +478,7 @@ class SecleaAI:
 
         # validate the splits? maybe later when we have proper Dataset class to manage these things.
         dataset_pks = [
-            str(hash(pd.util.hash_pandas_object(dataset).sum() + self._project))
+            dataset_hash(dataset, self._project)
             for dataset in [train_dataset, test_dataset, val_dataset]
             if dataset is not None
         ]
@@ -416,6 +526,7 @@ class SecleaAI:
         self._upload_model_state(
             model=model,
             training_run_pk=self._training_run,
+            dataset_pks=dataset_pks,
             sequence_num=0,
             final=True,
             model_manager=framework,
@@ -455,7 +566,7 @@ class SecleaAI:
                     url_path="/collection/projects", query_params={"name": project_name}
                 )
                 resp = handle_response(
-                    resp, 200, f"There was an issue getting the project: {resp.text}"
+                    resp, 200, f"There was an issue getting the project: {resp.reason}"
                 )
                 self._project = resp.json()[0]["id"]
 
@@ -472,7 +583,7 @@ class SecleaAI:
             query_params={"name": project_name, "organization": self._organization},
         )
         handle_response(
-            project_res, 200, f"There was an issue getting the projects: {project_res.text}"
+            project_res, 200, f"There was an issue getting the projects: {project_res.reason}"
         )
         if len(project_res.json()) == 0:
             return None
@@ -501,7 +612,7 @@ class SecleaAI:
             query_params={"organization": self._organization},
         )
         return handle_response(
-            res, expected=201, msg=f"There was an issue creating the project: {res.text}"
+            res, expected=201, msg=f"There was an issue creating the project: {res.reason}"
         )
 
     def _set_model(self, model_name: str, framework: ModelManagers) -> int:
@@ -553,6 +664,34 @@ class SecleaAI:
             model_pk = resp.json()[0]["id"]
         return model_pk
 
+    @staticmethod
+    def _ensure_required_metadata(metadata: Dict, defaults_spec: Dict) -> Dict:
+        """
+        Ensures that required metadata that can be specified by the user are filled.
+        @param metadata: The metadata dict
+        @param defaults_spec:
+        @return: metadata
+        """
+        for required_key, default in defaults_spec.items():
+            try:
+                if metadata[required_key] is None:
+                    metadata[required_key] = default
+            except KeyError:
+                metadata[required_key] = default
+        return metadata
+
+    @staticmethod
+    def _add_required_metadata(metadata: Dict, required_spec: Dict) -> Dict:
+        """
+        Adds required - non user specified fields to the metadata
+        @param metadata: The metadata dict
+        @param required_spec:
+        @return: metadata
+        """
+        for required_key, default in required_spec.items():
+            metadata[required_key] = default
+        return metadata
+
     def _upload_dataset(
         self,
         dataset: DataFrame,
@@ -561,77 +700,40 @@ class SecleaAI:
         parent_hash: Union[int, None],
         transformation: Union[DatasetTransformation, None],
     ):
-        split = transformation.split if transformation is not None else None
-        if parent_hash is not None:
-            parent_hash = hash(parent_hash + self._project)
-            # check parent exists - throw an error if not.
-            res = self._transmission.get(
-                url_path=f"/collection/datasets/{parent_hash}",
-                query_params={"project": self._project, "organization": self._organization},
-            )
-            if not res.status_code == 200:
-                raise AssertionError(
-                    "Parent Dataset does not exist on the Platform. Please check your arguments and "
-                    "that you have uploaded the parent dataset already"
-                )
-            parent = res.json()
-            # deal with the splits - take the set one by default but inherit from parent if None
-            if transformation.split is None:
-                # check the parent split - inherit split
-                split = parent["metadata"]["split"]
-            try:
-                if metadata["outcome_name"] is None:
-                    pass
-            except KeyError:
-                try:
-                    metadata["outcome_name"] = parent["metadata"]["outcome_name"]
-                except KeyError:
-                    metadata["outcome_name"] = None
-
-        # this needs to be here so split is always set.
-        metadata = {**metadata, "split": split, "features": list(dataset.columns)}
-
-        # ensure there is always an index.
-        try:
-            if metadata["index"] is None:
-                metadata["index"] = 0
-        except KeyError:
-            metadata["index"] = 0
-
         # upload a dataset - only works for a single transformation.
         if not os.path.exists(self._cache_dir):
             os.makedirs(self._cache_dir)
 
-        dataset_path = os.path.join(self._cache_dir, "tmp.csv")
-        dataset.to_csv(dataset_path, index=True)
-        comp_path = os.path.join(self._cache_dir, "compressed")
-        rb = open(dataset_path, "rb")
-        comp_path = save_object(rb, comp_path, compression=CompressionFactory.ZSTD)
-
-        dataset_hash = hash(pd.util.hash_pandas_object(dataset).sum() + self._project)
+        # TODO refactor to make multithreading safe.
+        dataset_file_path = save_dataset(
+            dataset, dataset_name, os.path.join(self._cache_dir, uuid.uuid4().__str__())
+        )
+        dset_pk = dataset_hash(dataset, self._project)
 
         response = post_dataset(
             transmission=self._transmission,
-            dataset_file_path=comp_path,
+            dataset_file_path=dataset_file_path,
             project_pk=self._project,
             organization_pk=self._organization,
             name=dataset_name,
             metadata=metadata,
-            dataset_hash=str(dataset_hash),
+            dataset_pk=dset_pk,
             parent_dataset_hash=str(parent_hash) if parent_hash is not None else None,
-            delete=True,
         )
         handle_response(
-            response, 201, f"There was some issue uploading the dataset: {response.text}"
+            response, 201, f"There was some issue uploading the dataset: {response.reason}"
         )
+        # tidy up files.
+        os.remove(dataset_file_path)
 
         # upload the transformations
-        if response.status_code == 201 and transformation is not None:
-            # upload transformations.
-            self._upload_transformation(
-                transformation=transformation,
-                dataset_pk=str(dataset_hash),
-            )
+        if response.status_code == 201:
+            if transformation is not None:
+                # upload transformations.
+                self._upload_transformation(
+                    transformation=transformation,
+                    dataset_pk=dset_pk,
+                )
 
     def _upload_model(self, model_name: str, framework: ModelManagers):
         """
@@ -651,7 +753,7 @@ class SecleaAI:
             query_params={"organization": self._organization, "project": self._project},
         )
         return handle_response(
-            res, expected=201, msg=f"There was an issue uploading the model: {res.text}"
+            res, expected=201, msg=f"There was an issue uploading the model: {res.reason}"
         )
 
     def _upload_training_run(
@@ -678,13 +780,14 @@ class SecleaAI:
             query_params={"organization": self._organization, "project": self._project},
         )
         return handle_response(
-            res, expected=201, msg=f"There was an issue uploading the training run: {res.text}"
+            res, expected=201, msg=f"There was an issue uploading the training run: {res.reason}"
         )
 
     def _upload_model_state(
         self,
         model,
         training_run_pk: int,
+        dataset_pks: List[int],
         sequence_num: int,
         final: bool,
         model_manager: ModelManagers,
@@ -694,10 +797,11 @@ class SecleaAI:
             exist_ok=True,
         )
         model_data = serialize(model, model_manager)
-        save_path = os.path.join(
-            Path.home(), f".seclea/{self._project_name}/{training_run_pk}/model-{sequence_num}"
+        file_name = f"data-{dataset_pks[0]}-model-{sequence_num}"
+        save_path = os.path.join(Path.home(), f".seclea/{self._project_name}/{training_run_pk}")
+        save_path = save_object(
+            model_data, file_name, save_path, compression=CompressionFactory.ZSTD
         )
-        save_path = save_object(model_data, save_path, compression=CompressionFactory.ZSTD)
 
         res = post_model_state(
             self._transmission,
@@ -707,11 +811,13 @@ class SecleaAI:
             str(training_run_pk),
             sequence_num,
             final,
-            True,
         )
 
+        # tidy up files.
+        os.remove(save_path)
+
         res = handle_response(
-            res, expected=201, msg=f"There was an issue uploading a model state: {res}"
+            res, expected=201, msg=f"There was an issue uploading a model state: {res.reason}"
         )
         return res
 
@@ -733,7 +839,7 @@ class SecleaAI:
         res = handle_response(
             res,
             expected=201,
-            msg=f"There was an issue uploading the transformations on transformation {idx} with name {transformation.func.__name__}: {res.text}",
+            msg=f"There was an issue uploading the transformations on transformation {idx} with name {transformation.func.__name__}: {res.reason}",
         )
         return res
 
@@ -747,7 +853,7 @@ class SecleaAI:
             query_params={"training_run": training_run_pk},
         )
         res = handle_response(
-            res, expected=200, msg=f"There was an issue loading the transformations: {res.text}"
+            res, expected=200, msg=f"There was an issue loading the transformations: {res.reason}"
         )
         transformations = list(map(lambda x: x["code_encoded"], res.json()))
         return list(map(decode_func, transformations))
@@ -774,7 +880,20 @@ class SecleaAI:
             return ModelManagers.XGBOOST
         elif "lightgbm" in module:
             return ModelManagers.LIGHTGBM
+        # TODO improve to exclude other keras backends...
+        elif "tensorflow" in module or "keras" in module:
+            return ModelManagers.TENSORFLOW
         elif "sklearn" in module:
             return ModelManagers.SKLEARN
         else:
             return ModelManagers.NOT_IMPORTED
+
+    @staticmethod
+    def _get_dataset_type(dataset: DataFrame) -> str:
+        if not np.issubdtype(dataset.index.dtype, np.integer):
+            try:
+                pd.to_datetime(dataset.index.values)
+            except (ParserError, ValueError):  # Can't cnvrt some
+                return "tabular"
+            return "time_series"
+        return "tabular"
